
#include <stdlib.h>
#include <stdio.h>
#include <string.h>

#include <iostream>
#include <string>
#include <sstream>
#include <set>

#include "CAutomaton.h"

using namespace std;

CAutomaton::CAutomaton(void)
{

}

CAutomaton::~CAutomaton(void)
{
	Release();
}

void CAutomaton::Release()
{
	m_end.clear();

	for(list<int*>::const_iterator it = m_func.begin(); it != m_func.end(); it++)
	{
		free(*it);
	}

	m_func.clear();
}

void CAutomaton::Copy(const CAutomaton& from)
{
	Release();

	m_init = from.m_init;
	m_alphbt = from.m_alphbt; 
	m_size = from.m_size;
			
	for(list<int>::const_iterator it = from.m_end.begin(); it != from.m_end.end(); it++)
	{
		m_end.push_back(*it);
	}

	for(list<int*>::const_iterator it = from.m_func.begin(); it != from.m_func.end(); it++)
	{
		int* st = (int*)malloc(sizeof(int) * m_alphbt);
		memcpy(st, *it, sizeof(int) * m_alphbt);
		m_func.push_back(st);
	}
}

//dohledam vsechny spoje, ktere jdou do "state" a pridam je do "to"
void CAutomaton::AddWays(int state, list<Tway*>& to, int max_state)
{	
	int pos = 0;
	for(list<int*>::const_iterator it = m_func.begin(); it != m_func.end() && pos < max_state; it++)
	{
		for(int i = 0; i < m_alphbt; i++)
		{
			if ((*it)[i] == state)
			{
				Tway* w = new Tway();
				w->alpha = i;
				w->state = pos + 1;
				w->func = *it;
				to.push_back(w);
			}
		}

		pos++;
	}
}

void CAutomaton::Dualize(CAutomaton& dual)
{
	//zkopiruji jeden automat do druheho
	dual.Copy(*this);

	/*
	projdu z koncovych stavu celou cestu jak se k nim lze dostat
	a kazdy spoj si ulozim
	*/
	list<Tway*> todual;
	set<int> dupl;

	//pridam vsechny koncove stavy
	for(list<int>::const_iterator it = dual.m_end.begin(); it != dual.m_end.end(); it++)
	{
		dupl.insert(*it);
		dual.AddWays(*it, todual, m_size);
	}
			
	//jdu dokud to jde
	pair<set<int>::iterator,bool> ret;
	for(list<Tway*>::const_iterator it = todual.begin(); it != todual.end(); it++)
	{
		ret = dupl.insert((*it)->state); //kontroluji duplicity
		if (ret.second)
			dual.AddWays((*it)->state, todual, m_size);
	}

	//Cislo dalsiho stavu
	int next = dual.m_func.size() + 1;

	//vytvorim si "peklo"
	int hell = next++;
	int* hellf = (int*)malloc(sizeof(int) * m_alphbt);
	for(int i = 0; i < m_alphbt; i++)
		hellf[i] = hell;
	dual.m_func.push_back(hellf);

	//kazdy stav rozdelim na dva po sobe jdouci, vsechny vazby z druheho stavu jdou do pekla, mimo puvodni spoj
	for(list<Tway*>::const_iterator it = todual.begin(); it != todual.end(); it++)
	{
		int* row = (*it)->func;
		int* news = (int*)malloc(sizeof(int) * m_alphbt);

		memcpy(news, hellf, sizeof(int) * m_alphbt);

		news[(*it)->alpha] = row[(*it)->alpha];
		row[(*it)->alpha] = next;

		dual.m_func.push_back(news);

		next++;

		delete(*it);			
	}
	
	dual.m_size = dual.m_func.size();
}

ostream& operator <<(ostream& out, const CAutomaton& a)
{
	out << a.m_size << endl;
	out << a.m_alphbt << endl;
	out << a.m_init << endl;
	
	list<int>::const_iterator it = a.m_end.begin();
	while(it != a.m_end.end())
	{
		out << *it << " ";
		it++;
	}
	out << endl;

	for(list<int*>::const_iterator it = a.m_func.begin(); it != a.m_func.end(); it++)
	{
		for(int j = 0; j < a.m_alphbt; j++)
		{
			out << (*it)[j] << " ";
		}
		out << endl;
	}

	return out;
}
istream& operator >>(istream& in, CAutomaton& a)
{
	a.Release();

	in >> a.m_size;
	in >> a.m_alphbt;
	in >> a.m_init;

	if (!in.good())
		return in;
	
	while(in.peek() != '\n')
		in.ignore(1);
	in.ignore(1);

	int temp;
	char buffer[255];
	
	in.getline(buffer, 255);
	istringstream str(buffer, istringstream::in);
		
	while(1)
	{
		str >> temp;
		if (str.fail())
			break;

		a.m_end.push_back(temp);
	}

	if (!in.good())
		return in;

	for(int i = 0; i < a.m_size; i++)
	{
		int* fc = (int*)malloc(sizeof(int) * a.m_alphbt);
		a.m_func.push_back(fc);

		for(int j = 0; j < a.m_alphbt; j++)
		{
			in >> temp;
			fc[j] = temp;
		}
	}


	return in;
}
